#include <stdio.h>
#include <math.h>
#include <float.h>
#include <stdlib.h>
#include "arithcommon.h"
#include "gpubench.h"
#include "getopt.h"
#include <GL/glu.h>

char const * shader_ati();
char const * shader_nv();
void check_results(float * results, float * xData, float * yData, float * zData, int arraySize);


static PFNGLACTIVETEXTUREARBPROC       glActiveTextureARB;
static PFNGLMULTITEXCOORD2FARBPROC     glMultiTexCoord2fARB;
static PFNGLMULTITEXCOORD4FVARBPROC     glMultiTexCoord4fvARB;

enum TestType
{
	Specials,
	AddOvf,
	MadGuardBits,
	MadRange,
	Add1GuardBits,
	Add2GuardBits,
	SignNotation,
	MultTrunc,
	MultError,
	MultShift
};

static int nvtype[5] = {0,
			GL_FLOAT_R32_NV,
			GL_FLOAT_RG32_NV,
			GL_FLOAT_RGB32_NV,
			GL_FLOAT_RGBA32_NV};
static int atitype[5] = {0,
			 GL_ALPHA_FLOAT32_ATI,
			 GL_LUMINANCE_ALPHA_FLOAT32_ATI,
			 GL_RGB_FLOAT32_ATI,
			 GL_RGBA_FLOAT32_ATI};

static int format[5] = {0, GL_RED, GL_LUMINANCE_ALPHA, 
			GL_RGB, GL_RGBA};


static void fill_data_ovf(float * xData, float * yData, float * zData, int components, int size)
{
  int arraySize = size * size * components;
  for(int i=0; i<arraySize; i+=4) {
    xData[i] = xData[i+1] = xData[i+2] = xData[i+3] = FLT_MAX;
    yData[i] = yData[i+1] = yData[i+2] = yData[i+3] = FLT_MAX;
    zData[i] = zData[i+1] = zData[i+2] = zData[i+3] = FLT_MAX;
  }
}

static void fill_data_mad_guard_bits(float * xData, float * yData, float * zData, int components, int size)
{
    int arraySize = size * size * components;
    for(int i=0; i<arraySize; i++) {
	xData[i] = random_float(0, 0, 0);
	yData[i] = random_float(0, 0, 0);
	//zData[i] = - xData[i]*yData[i];
	zData[i] = 0.0f;
    }
}

static void fill_data_add1_guard_bits(float * xData, float * yData, float * zData, int components, int size)
{
	int arraySize = size * size * components;
	float tpmi = -0.5f;
	for (int i=0;i<arraySize;i++)
	{
		xData[i] = 1.5f;
		yData[i] = tpmi;
   		zData[i] = 0.0f; // not used
		tpmi *= .5f;
	}
}

static void fill_data_add2_guard_bits(float * xData, float * yData, float * zData, int components, int size)
{
    int arraySize = size * size * components;
    float tpmi = -.5f;
    for (int i=0;i<arraySize;i++)
    {
		xData[i] = 1.0f;
		yData[i] = 0.5f;
		zData[i] = tpmi;
		tpmi *= .5f;
    }
}

static void fill_data_sign_notation(float * xData, float * yData, float * zData, int components, int size)
{
    int arraySize = size * size * components;
    for(int i=0; i<arraySize; i+=4)
    {
	xData[i] = random_float(0, 0, 0);
	xData[i+1] = - xData[i];
	xData[i+2] = - xData[i];
	xData[i+3] = xData[i];

	yData[i] = random_float(0, 0, 0);
	yData[i+1] = - yData[i];
	yData[i+2] = yData[i];
	yData[i+3] = - yData[i];
	

	zData[i] = zData[i+1] = zData[i+2] = zData[i+3] = 0.0f;
    }
}

static void fill_data_mult_error(float * xData, float * yData, float * zData, int components, int size)
{
    int arraySize = size * size * components;
    for(int i=0; i<arraySize; i+=2) {

	xData[i] = random_float(0, 0, 0);
	yData[i] = random_float_interval(1.0f, 2.0f / xData[i]);
	//yData[i] = (2.0f - ldexp(1.0f, -23)) / xData[i];
	zData[i] = 0;

	xData[i+1] = xData[i];
	yData[i+1] = random_float_interval(2.0f / xData[i+1], 2.0f);
    }
}

static void fill_data_mult_trunc(float * xData, float * yData, float * zData, int components, int size)
{
	int arraySize = size * size * components;
	float step = 1.0f / arraySize;
	
    float x = 1.0f + (float)ldexp(1.0, -23);
    for (int i=0;i<arraySize;i++)
    {

	xData[i] = x;
	yData[i] = 1.0f + i * step;
	zData[i] = 0.0f; // not used
    }
}


static void fill_data_specials(float * xData, float * yData, float * zData, int components, int size)
{
	int const data_size = 14;
	float const zero = 0.0f;  // Workaround: VC++ insists that 1.0/0.0 is invalid

	static float input_data[data_size][3] = {
		{ -0.0f, 0.0f, -0.0f },
		{ FLT_MAX, 1.0f, 0.0f },      // big number
		{ FLT_MAX*2.0f, 1.0f, 0.0f },  // +inf propogation
		{ 1.0f/zero, 1.0f, 0.0f },  // -inf propagation
		{ FLT_MAX, 2.0f, 0.0f },       // mul overflow +
		{ -FLT_MAX, 2.0f, 0.0f },     // mul overflow -
		{ FLT_MAX, 1.0f, FLT_MAX },   // add overflow +
		{ -FLT_MAX, 1.0f, -FLT_MAX }, // add overflow -
		{ FLT_MIN/2.0f, 1.0f, 0.0f },     // denormal + propagation
		{ -FLT_MIN/2.0f, 1.0f, 0.0f },    // denormal - propagation
		{ FLT_MIN, .5f, 0.0f },  // denormal + generation
		{ -FLT_MIN, .5f, 0.0f }, // denormal - generation
		{ (float)sqrt(-1.0f), 1.0f, 0.0f },  // NaN propagation
		{ FLT_MAX*2.0f, 0.0f, 0.0f }};  // NaN generation
		

    int arraySize = size * size * components;
    for(int i=0; i<arraySize; i++) {
		if(i < data_size)
		{
			xData[i] = input_data[i][0];
			yData[i] = input_data[i][1];
			zData[i] = input_data[i][2];
		}
		else
		{
			xData[i] = 0.0f;
			yData[i] = 0.0f;
			zData[i] = 0.0f;
		}
    }
}

static void fill_data_mad_range(float * xData, float * yData, float * zData, int components, int size)
{
	int arraySize = size * size * components;
	float bignum = FLT_MAX;
	float factor = 1.0f;
	for(int i=0; i<arraySize; i++) {
		xData[i] = bignum;
		yData[i] = factor;
		zData[i] = - bignum;
	
		factor += 1/16.0f;
	}
}

static void fill_data_mult_shift(float * xData, float * yData, float * zData, int components, int size)
{
    int arraySize = size * size * components;
    for(int i=0; i<arraySize; i+=2)
    {
	xData[i] = random_float(0, 0, 0);
	xData[i+1] = 2 * xData[i];

	yData[i] = yData[i+1] = random_float(0, 0, 0);

	zData[i] = zData[i+1] = 0.0f;
    }
}


static void fill_data(float * xData, float * yData, float * zData, int components, int size, enum TestType test_type)
{
	switch(test_type)
	{
	case AddOvf:
		fill_data_ovf(xData, yData, zData, components, size);
		break;
	case MadGuardBits:
		fill_data_mad_guard_bits(xData, yData, zData, components, size);
		break;
	case Add1GuardBits:
		fill_data_add1_guard_bits(xData, yData, zData, components, size);
		break;
	case Add2GuardBits:
		fill_data_add2_guard_bits(xData, yData, zData, components, size);
		break;
	case SignNotation:
		fill_data_sign_notation(xData, yData, zData, components, size);
		break;
	case MultError:
		fill_data_mult_error(xData, yData, zData, components, size);
		break;
	case MultTrunc:
		fill_data_mult_trunc(xData, yData, zData, components, size);
		break;
	case Specials:
		fill_data_specials(xData, yData, zData, components, size);
		break;
	case MadRange:
		fill_data_mad_range(xData, yData, zData, components, size);
		break;
	case MultShift:
		fill_data_mult_shift(xData, yData, zData, components, size);
		break;
	
	}
}


static void
generate_shader(char* fpStr,
                  TestType test_type, int components, shader_type shader)
{

	if(shader == PixelShader)
	{
		sprintf(fpStr,
			"!!ARBfp1.0\n"
			"TEMP x;\n"
			"TEMP y;\n"
			"TEMP z;\n"
			"TEMP tmp;\n"
			"TEMP res;\n");
			
		if(gpubench_isNV())
		{
			strcat(fpStr,
				"TEX x, fragment.texcoord[0], texture[0], RECT;\n"
				"TEX y, fragment.texcoord[0], texture[1], RECT;\n"
				"TEX z, fragment.texcoord[0], texture[2], RECT;\n");
		}
		else
		{
			strcat(fpStr,
				"TEX x, fragment.texcoord[0], texture[0], 2D;\n"
				"TEX y, fragment.texcoord[0], texture[1], 2D;\n"
				"TEX z, fragment.texcoord[0], texture[2], 2D;\n");
		}
	}
	else
	{
		sprintf(fpStr,
			"!!ARBvp1.0\n"
			"ATTRIB pos = vertex.position;\n"
			"ATTRIB x = vertex.texcoord[0];\n"
			"ATTRIB y = vertex.texcoord[1];\n"
			"ATTRIB z = vertex.texcoord[2];\n"
			"TEMP tmp;\n"
			"TEMP res;\n"
			"TEMP xfNormal;\n"
			"PARAM mat[4] = { state.matrix.mvp };\n"
			"PARAM mvinv[4] = {state.matrix.modelview.invtrans };\n"

			"# vertex position -> clip space\n"
			"DP4 res.w, pos, mat[3];\n"
			"DP4 res.x, pos, mat[0];\n"
			"DP4 res.y, pos, mat[1];\n"
			"DP4 res.z, pos, mat[2];\n"
			"RCP tmp.w, res.w;\n"
			"MUL res, res, tmp.w;\n"
			"MOV result.position, res;\n");
	}


	switch(test_type)
	{
	case AddOvf:
		strcat(fpStr,
			"ADD tmp, x, y;\n"
			"SUB res, tmp, z;\n");
		break;
	case MadGuardBits:
		strcat(fpStr,
			"MUL tmp, x, y;\n"
			"MAD res, x, y, -tmp;\n");
		break;
	case Add1GuardBits:
		strcat(fpStr,
			"ADD res, x, y;\n");
		break;
	case Add2GuardBits:
		strcat(fpStr,
			"ADD tmp, x, y;\n"
			"ADD res, tmp, z;\n");
		break;
	case SignNotation:
	case MultError:
	case MultTrunc:
	case MultShift:
		strcat(fpStr,
			"MUL res, x, y;\n");
		break;
	case Specials:
	case MadRange:
		strcat(fpStr,
			"MAD res, x, y, z;\n");
		break;
	}

	if(shader == PixelShader)
	{
		strcat(fpStr, "MOV result.color, res;\n");
	}
	else
	{
		strcat(fpStr, "MOV result.texcoord, res;\n");
	}
	strcat(fpStr, "END\n");

}

static int testSize(TestType test_type)
{
	switch(test_type)
	{
	case AddOvf:
		return 1;
		break;
	case Specials:
	case MadRange:
		return 2;
		break;
	case Add1GuardBits:
	case Add2GuardBits:
		return 4;
		break;
	case MadGuardBits:
	case SignNotation:
	case MultError:
	case MultTrunc:
	case MultShift:
		return 1024;
		break;
	}
	return 1024;
}

static void check_results_mad_guard_bits(float * results, float * xData, float * yData, float * zData, int arraySize)
{
    double dmax=-1;
    for(int i = 0; i < arraySize; i++)
    {
	double diff = abs(results[i]);

	if(diff > dmax)
	{
	    dmax = diff;
	}
    }

    printf("abs(MAD(x, y, -x*y)) <= %.12f\n", dmax, dmax);

}


static void check_results_add1_guard_bits(float * results, float * xData, float * yData, float * zData, int arraySize)
{
	// check how many bit are used within the adder :
	printf("Computing r = 1.5 - 2^(-i) (first mad)\n");
	for(int i = 0; i < arraySize; i++)
	{
		printf("i=%d ->\tr=%a\t(%.8f)\n", i+1, results[i], results[i]);
	}

}

static void check_results_add2_guard_bits(float * results, float * xData, float * yData, float * zData, int arraySize)
{
    // check how many bit are used within the adder :
    //printf("Computing r = (1.5 - 2^(-i)) - 1.5  (second mad)\n");
	printf("Computing r = 1.5 - 2^(-i) (second mad)\n");
	for(int i = 0; i < arraySize; i++)
    {
		printf("i=%d ->\tr=%a\t(%.8f)\n", i+1, results[i], results[i]);
    }
}

static void check_results_sign_notation(float * results, float * xData, float * yData, float * zData, int arraySize)
{
    double dmax[3] = {-1.0, -1.0, -1.0};
    for(int i = 0; i < arraySize; i += 4)
    {
	double diff = abs(results[i+1] - results[i]);
	if(diff > dmax[0]) {
	    dmax[0] = diff;
	}
	for(int j = 1; j < 3; j++)
	{
	    diff = abs(results[i+j+1] + results[i]);
	    if(diff > dmax[j]) {
		dmax[j] = diff;
	    }
	}
    }
    
    printf("abs(x*y + (-x)*y) <= %.12f\n", dmax[1]);
    printf("abs(x*y + x*(-y)) <= %.12f\n", dmax[2]);
    printf("abs(x*y - (-x)*(-y)) <= %.12f\n", dmax[0]);

}

static void check_results_mult_error(float * results, float * xData, float * yData, float * zData, int arraySize)
{
    double emin[2] = {1.0, 1.0}, emax[2] = {-1.0, -1.0}, esum[2] = {0.0, 0.0};
    
	float xmin[2], xmax[2], ymin[2], ymax[2], zmin[2], zmax[2];

    for(int i = 0; i < arraySize; i++)
    {
	//printf("x=%f, y=%f,, x*y=%f, r=%f\n", xData[i], yData[i], xData[i] * yData[i], results[i]);
		double result_nearest = (double)xData[i] * (double)yData[i];
		// truncate
	double err = (double)results[i] - result_nearest;

	if(err > emax[i%2]) {
	    emax[i%2] = err;
		xmax[i%2] = xData[i];
		ymax[i%2] = yData[i];
		zmax[i%2] = results[i];
	}
//	if(err < emin[i%2] && results[i] == 2.0f-ldexp(1.0f, -22)) {
	if(err < emin[i%2]) {
	    emin[i%2] = err;
		xmin[i%2] = xData[i];
		ymin[i%2] = yData[i];
		zmin[i%2] = results[i];
	}
	esum[i%2] += err;
    }

    double eavg0 = esum[0] / arraySize;
    double eavg1 = esum[1] / arraySize;

	float ulp1 = ldexp(1.0f, -23);
	float ulp2 = ldexp(1.0f, -22);

    printf("Multiplication error:\n");
    printf(" x in [1,2), y in [1, 2/x) -> min=%a, max=%a, avg=%a\n", emin[0], emax[0], eavg0);
	printf("                             (min=%f ulp, max=%f ulp)\n", emin[0] / ulp1, emax[0] / ulp1);
    printf(" x in [1,2), y in [2/x, 2) -> min=%a, max=%a, avg=%a\n", emin[1], emax[1], eavg1);
	printf("                             (min=%f ulp, max=%f ulp)\n", emin[1] / ulp2, emax[1] / ulp2);

	printf("y in [1, 2/x), min for x=%a, y=%a, r=%a\n", xmin[0], ymin[0], zmin[0]);
	printf("               max for x=%a, y=%a, r=%a\n", xmax[0], ymax[0], zmax[0]);

	printf("y in [2/x, 2), min for x=%a, y=%a, r=%a\n", xmin[1], ymin[1], zmin[1]);
	printf("               max for x=%a, y=%a, r=%a\n", xmax[1], ymax[1], zmax[1]);


}

static void check_results_mult_trunc(float * results, float * xData, float * yData, float * zData, int arraySize)
{
    int k = -1;
	float step = 1.0f / arraySize;
    for(int i = 0; i < arraySize; i++)
    {
	if(results[i] != 1.0f + i * step + (float)ldexp(1.0, -23))
	{
	    k = i;
	    break;
	}
    }

	if(k != -1) {
		printf("k=%d x %a (0x%x), x=%a, y=%a, r=%a\n", k, step, k, xData[k], yData[k], results[k]);
	}
	else {
		printf("No truncation???\n");
	}


}

static void check_results_mad(float * results, float * xData, float * yData, float * zData, int arraySize)
{
    for(int i = 0; i < arraySize; i++)
    {
	printf("%a * %a + %a = %a\n", xData[i], yData[i], zData[i], results[i]);
    }
}

static void check_results_mult_shift(float * results, float * xData, float * yData, float * zData, int arraySize)
{
    double dmax = -1.0;
    for(int i = 0; i < arraySize; i += 2)
    {
	double diff = abs(results[i] - results[i+1]/2);
	if(diff > dmax) {
	    dmax = diff;
	}
    }

    printf("abs(x*y + (2x)*y / 2) <= %.12f\n", dmax);


}

static void check_results(float * results, float * xData, float * yData, float * zData,
	int arraySize, TestType test_type)
{
	switch(test_type)
	{
	case AddOvf:
		printf("(M+M)-M returned %f\n", results[0]);
		break;
	case MadGuardBits:
		check_results_mad_guard_bits(results, xData, yData, zData, arraySize);
		break;
	case Add1GuardBits:
		check_results_add1_guard_bits(results, xData, yData, zData, arraySize);
		break;
	case Add2GuardBits:
		check_results_add2_guard_bits(results, xData, yData, zData, arraySize);
		break;
	case SignNotation:
		check_results_sign_notation(results, xData, yData, zData, arraySize);
		break;
	case MultError:
		check_results_mult_error(results, xData, yData, zData, arraySize);
		break;
	case MultTrunc:
		check_results_mult_trunc(results, xData, yData, zData, arraySize);
		break;
	case Specials:
	case MadRange:
		check_results_mad(results, xData, yData, zData, arraySize);
		break;
	case MultShift:
		check_results_mult_shift(results, xData, yData, zData, arraySize);
		break;
	}
	
}

void print_out(const char *name, float * data, float * dataX, int N){
    FILE * F;
    long long int i;

    F = fopen(name, "w");

    for(i=0; i<N; i+=4)
	fprintf(F,"%e %e %a %a \t %a %a \t %a %a \t %a %a \n",data[i],   dataX[i], data[i],   dataX[i],
		data[i+1], dataX[i+1],
		data[i+2], dataX[i+2],
		data[i+3], dataX[i+3]);

    fclose(F);
}


void usage(char *name) {

    fprintf (stderr, "Usage: %s <options>\n", name);
    fprintf (stderr, "  Options\n");
    fprintf (stderr, "  -s, --size=SIZE\n");
    fprintf (stderr, "             size of the texture and framebuffer (512)\n");
    fprintf (stderr, "  -c, --components=SIZE\n");
    fprintf (stderr, "            number of components for texture and\n");
    fprintf (stderr, "            render target. (4)\n");
    fprintf (stderr, "  -i, --iterations=INT\n");
    fprintf (stderr, "            Number of iterations to perform (default: 1000)\n");
    fprintf (stderr, "  -t, --test=STRING\n");
    fprintf (stderr, "            Specifies which arithmetic test to perform.\n");
    fprintf (stderr, "            Can be Specials, AddOvf, MadGuardBits, MadRange, Add1GuardBits,\n");
    fprintf (stderr, "            Add2GuardBits, SignNotation, MultShift, MultError, MultTrunc.\n");
    fprintf (stderr, "  -u, --unit=STRING (pixel)\n");
    fprintf (stderr, "            Shader unit to test.\n");
    fprintf (stderr, "            Can be pixel, vertex.\n");
    exit(0);
}

static int
dotest(int size,
       int components, int iters, TestType test_type, shader_type shader,
	   float * xData, float * yData, float * zData)
{
    shader_t multFP;
    shader_t nop;
	int i;
	
	// 100K should be enough ;-)
	char * fpStr = (char*)malloc(1024*100);
	
	generate_shader(fpStr, test_type, components, shader);

	//printf(fpStr);
	
    if(shader == PixelShader) {
		multFP = gpubench_loadShaderFromString(fpStr);
    }
    else {
		multFP = gpubench_loadVertexShaderFromString(fpStr);
		nop = gpubench_loadShader(gpubench_isNV() ? "shaders/nop/nop_nv.arb" : "shaders/nop/nop_ati.arb");
    }

    if (multFP == GPUBENCH_BOGUS_PROGRAM) {
		fprintf(stderr, "Error, couldn't load program.\n");
		gpubench_destroyPbuffer();
		exit(1);
    }



    if(shader == PixelShader) {
		gpubench_bindShader(multFP);
    }
    else {
		gpubench_bindVertexShader(multFP);
		gpubench_bindShader(nop);
    }
	
    int draw_to_front = 1;

    float scaleX = (gpubench_isNV()) ? 1.0f : (float)size;
    float scaleY = (gpubench_isNV()) ? 1.0f : (float)size;

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    // force load of textures onto card by rendering geometry
    glBegin(GL_TRIANGLES);    
    glTexCoord2f(0.0f, size*2.0f / scaleY);      
    glVertex2f(-1.0f, 3.0f); 
    glTexCoord2f(0.0f, 0.0f);      
    glVertex2f(-1.0f, -1.0f);   
    glTexCoord2f(2.0f*size / scaleX, 0.0f);      
    glVertex2f(3.0f, -1.0f);
    glEnd();

    glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    gpubench_resetTimer();


    for (i=0;i<iters; i++) {

	if(shader == PixelShader)
	{
	    glBegin(GL_TRIANGLES);    
	    glTexCoord2f(0.0f, size*2.0f / scaleY);      
	    glVertex2f(-1.0f, 3.0f); 
	    glTexCoord2f(0.0f, 0.0f);      
	    glVertex2f(-1.0f, -1.0f);   
	    glTexCoord2f(2.0f*size / scaleX, 0.0f);      
	    glVertex2f(3.0f, -1.0f);
	    glEnd();
	}
	else
	{
	    float factor = 2.0f/size;
	    glBegin(GL_POINTS);
	    for(int y = 0; y < size; y++)
	    {
			float fy = -1.0f+(y+.5f)*factor;
			for(int x = 0; x < size; ++x)
			{
				glMultiTexCoord4fvARB(GL_TEXTURE0_ARB, &xData[(y*size+x)*4]);
				glMultiTexCoord4fvARB(GL_TEXTURE1_ARB, &yData[(y*size+x)*4]);
				glMultiTexCoord4fvARB(GL_TEXTURE2_ARB, &zData[(y*size+x)*4]);
				glVertex2f(-1.0f+(x+.5f)*factor, fy);
			}
	    }
	    glEnd();
	}

	draw_to_front = gpubench_bindAndSwap(draw_to_front);
    }
	
	free(fpStr);
	return draw_to_front;
}



int main(int argc, char** argv)
{
    char *progname = argv[0];

    int opt;
    int opt_index;

    int size = 1024;
    int iters = 1;
    int components = 4;
    bool doVerify = true;
	
	enum TestType test_type = AddOvf;
	enum shader_type shader = PixelShader;

	
    static struct option long_options[] = {
	{"size",         1, 0, 's'},
	{"components",   1, 0, 'c'},
	{"iters",        1, 0, 'i'},
	{"test",         1, 0, 't'},
	{"unit",         1, 0, 'u'},
	{"verify",       0, 0, 'z'},
	{0, 0, 0, 0}
    };

    while ((opt = getopt_long(argc, argv, "s:c:i:t:u:z",
			      long_options, &opt_index)) != EOF) {
	switch (opt) {
	case 's':
	    size = atoi(optarg);
	    break;
	case 'c':
	    components = atoi(optarg);
	    break;    
	case 'i':
	    iters = atoi(optarg);
	    break;
	case 't':
	{
		if(!strcmp(optarg, "AddOvf")) {
			test_type = AddOvf;
		}
		else if(!strcmp(optarg, "Specials")) {
			test_type = Specials;
		}
		else if(!strcmp(optarg, "MadGuardBits")) {
			test_type = MadGuardBits;
		}
		else if(!strcmp(optarg, "MadRange")) {
			test_type = MadRange;
		}
		else if(!strcmp(optarg, "Add1GuardBits")) {
			test_type = Add1GuardBits;
		}
		else if(!strcmp(optarg, "Add2GuardBits")) {
			test_type = Add2GuardBits;
		}
		else if(!strcmp(optarg, "SignNotation")) {
			test_type = SignNotation;
		}
		else if(!strcmp(optarg, "MultTrunc")) {
			test_type = MultTrunc;
		}
		else if(!strcmp(optarg, "MultError")) {
			test_type = MultError;
		}
		else if(!strcmp(optarg, "MultShift")) {
			test_type = MultShift;
		}
		else {
			usage(progname);
		}
		break;
	}
	case 'u':
		if(!strcmp(optarg, "pixel")) {
			shader = PixelShader;
		}
		else if(!strcmp(optarg, "vertex")) {
			shader = VertexShader;
		}
		else {
			usage(progname);
		}
		break;
	default:
	    usage(progname);
	}
    }
  
    argv += optind;
    argc -= optind;
    if (argc < 0) 
		usage(progname);
		
	size = testSize(test_type);
		
    if (size < 1 || size > 2048)      
		usage(progname);

		
    printf("(* size=%d components=%d iters=%d unit=%s *)\n", size, components, iters,
		shader == PixelShader? "pixel" : "vertex");
		
    fflush(stdout);

    gpubench_createWindow(10, 10, NULL, "Arithmetic test");
    GETPROCADDR(PFNGLACTIVETEXTUREARBPROC, glActiveTextureARB);
    GETPROCADDR(PFNGLMULTITEXCOORD2FARBPROC, glMultiTexCoord2fARB);  	
    GETPROCADDR(PFNGLMULTITEXCOORD4FVARBPROC, glMultiTexCoord4fvARB);
    gpubench_createFloatPbufferComponents(size, size, GL_TRUE, components);
    gpubench_makeCurrentPbuffer();

    glDrawBuffer(GL_BACK);
    glReadBuffer(GL_BACK);
    glDisable (GL_DEPTH_TEST);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glDrawBuffer(GL_FRONT);
    glReadBuffer(GL_FRONT);
    glDisable (GL_DEPTH_TEST);
    glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    int arraySize = components * size * size;
    float* xData = (float*)malloc(sizeof(float) * arraySize);
    float* yData = (float*)malloc(sizeof(float) * arraySize);
    float* zData = (float*)malloc(sizeof(float) * arraySize);

    fill_data(xData, yData, zData, components, size, test_type);

    // note: right now just using 2D textures on ATI since I don't
    // know how to read from RECT render texture on ATI.  This is 
    // unfortunate since we now have to create two shaders (one for 2D
    // textures and one for RECT textures)
    int textureType = gpubench_isNV() ? GL_TEXTURE_RECTANGLE_NV : GL_TEXTURE_2D;

    GLuint texId[3];

    glGenTextures(3, texId);

    glEnable(textureType);	
    glActiveTextureARB(GL_TEXTURE0_ARB);
    glBindTexture(textureType, texId[0]);
    glTexImage2D(textureType, 0, gpubench_isNV() ? nvtype[components] : atitype[components], size, size, 0,
		 format[components], GL_FLOAT, xData);
    glTexParameterf(textureType, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameterf(textureType, GL_TEXTURE_WRAP_T, GL_CLAMP);
    glTexParameterf(textureType, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameterf(textureType, GL_TEXTURE_MIN_FILTER, GL_NEAREST);

    CHECK_GL();

    glActiveTextureARB(GL_TEXTURE1_ARB);
    glBindTexture(textureType, texId[1]);
    glTexImage2D(textureType, 0, gpubench_isNV() ? nvtype[components] : atitype[components], size, size, 0,                
		 format[components], GL_FLOAT, yData);
    glTexParameterf(textureType, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameterf(textureType, GL_TEXTURE_WRAP_T, GL_CLAMP);
    glTexParameterf(textureType, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameterf(textureType, GL_TEXTURE_MIN_FILTER, GL_NEAREST);		

    CHECK_GL();    

    glActiveTextureARB(GL_TEXTURE2_ARB);
    glBindTexture(textureType, texId[2]);
    glTexImage2D(textureType, 0, gpubench_isNV() ? nvtype[components] : atitype[components], size, size, 0,                
		 format[components], GL_FLOAT, zData);
    glTexParameterf(textureType, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameterf(textureType, GL_TEXTURE_WRAP_T, GL_CLAMP);
    glTexParameterf(textureType, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameterf(textureType, GL_TEXTURE_MIN_FILTER, GL_NEAREST);		

    CHECK_GL();    
	
	int draw_to_front = dotest(size, components, iters, test_type, shader,
		xData, yData, zData);

    glFinish();
	
    float ms = gpubench_getTimer();

    CHECK_GL();

    if (doVerify) {
    
	draw_to_front = gpubench_bindAndSwap(draw_to_front);
  
	int numErrors = 0;
	float* results = (float*)malloc(sizeof(float) * arraySize);

	glReadPixels(0, 0, size, size, format[components], GL_FLOAT, results);

	check_results(results, xData, yData, zData, arraySize, test_type);


	free(results);

    }


    printf("%4.4f  msec per iter  (%4.4f msec total)\n", ms / 1000.0f, ms);
    printf("%4.4f  MFLOPS\n", (2.0f * iters * components * size * size) / ms / 1000.0f);

    free(xData);
    free(yData);
    free(zData);

    glDeleteTextures(3, texId);
    gpubench_destroyPbuffer();

    return 0;
}



