#ifdef WIN32
#include <windows.h>
#else
#define APIENTRY
#endif
#include <GL/gl.h>
#include <string.h>
#include <stdio.h>
#include <assert.h>

#include "gpubench.h"


#define GET_PROC(x) {					\
    void (**foo)();                                     \
    foo = (void (**)()) &x;                             \
    *foo = (void (*)()) wglGetProcAddress(#x);          \
    if (x == NULL) {					\
        fprintf(stderr,"Extension %s not found.\n", #x);\
                fflush(stderr);                         \
                abort();                                \
    }                                                   \
}


/* for ARB fragment programs */
PFNGLGENPROGRAMSARBPROC               glGenProgramsARB             = NULL;
PFNGLPROGRAMSTRINGARBPROC             glProgramStringARB           = NULL;
PFNGLBINDPROGRAMARBPROC               glBindProgramARB             = NULL;
PFNGLDELETEPROGRAMSARBPROC            glDeleteProgramsARB          = NULL;
PFNGLPROGRAMENVPARAMETER4FARBPROC     glProgramEnvParameter4fARB   = NULL;
PFNGLISPROGRAMARBPROC                 glIsProgramARB               = NULL;
PFNGLACTIVETEXTUREARBPROC             glActiveTextureARB           = NULL;
PFNGLCLIENTACTIVETEXTUREARBPROC       glClientActiveTextureARB     = NULL;


int load_program_string_arb (const char *source, GLenum shader_type) {
  int prog_id;
  GLuint prog_len;

  /* get function pointers if we don't have them already */
  if (!glGenProgramsARB) GET_PROC(glGenProgramsARB);
  if (!glProgramStringARB) GET_PROC(glProgramStringARB);

  /* generate program and read it in to the hardware */
  glGenProgramsARB(1, (GLuint *) &prog_id);

  if (!prog_id) {
    gpubench_setError(GPUBENCH_ERROR_SHADER_LOAD, "GenPrograms failed");
    return GPUBENCH_BOGUS_PROGRAM;
  }

  /* bind program */
  if (!glBindProgramARB) GET_PROC(glBindProgramARB);
  glBindProgramARB(shader_type, prog_id);

  /* read it in... */
  prog_len = strlen(source);
  glProgramStringARB(shader_type, GL_PROGRAM_FORMAT_ASCII_ARB, prog_len, (const GLubyte*) source);


  /* check if there were any errors with the fragment program */
  if (glGetError() != GL_NO_ERROR) {
    GLint ePos;
    char* error = (char*) glGetString(GL_PROGRAM_ERROR_STRING_ARB);
    glGetIntegerv(GL_PROGRAM_ERROR_POSITION_ARB, &ePos);
    /* taking this out helps perl digest the output without stalling */
    /*fprintf(stderr, "Shader error: %s\n", error);
    fprintf(stderr, "FSHADER::Load(): Error loading the program at location %d: %.30s\n",
            ePos, source + ePos);
    fflush(stderr);*/

    gpubench_setError(GPUBENCH_ERROR_SHADER_LOAD, "Parsing program");
    return GPUBENCH_BOGUS_PROGRAM;
  }

  /* make sure that the program has been loaded properly */
  if (!glIsProgramARB) GET_PROC(glIsProgramARB);
  if(!glIsProgramARB(prog_id)) {
    gpubench_setError(GPUBENCH_ERROR_SHADER_LOAD, "Invalid program id, something messed up");
    return GPUBENCH_BOGUS_PROGRAM;
  }

  return(prog_id);
}

int load_fragment_program_string_arb (const char * source) {
    return load_program_string_arb(source, GL_FRAGMENT_PROGRAM_ARB);
}

int load_vertex_program_string_arb (char const * source) {
    return load_program_string_arb(source, GL_VERTEX_PROGRAM_ARB);
}

int load_program_arb (const char *fname, GLenum shader_type) {
  FILE *fp;
  int n;
  int prog_id;
  GLuint prog_len;
  char *source;

  /* Open the file */
  fp = fopen(fname, "rb");

  if(!fp) {
    gpubench_setError(GPUBENCH_ERROR_SHADER_LOAD, "Couldn't open file");
    return GPUBENCH_BOGUS_PROGRAM;
  }

  /* Open the file and read the contents into the "source" string */
  fp = fopen(fname, "rb");
  assert (fp);
  fseek (fp, 0, SEEK_END);
  n = ftell(fp);
  rewind(fp);
  source = (char *) calloc (n+1, 1);
  fread (source, 1, n, fp);
  fclose(fp);

  /* get function pointers if we don't have them already */
  if (!glGenProgramsARB) GET_PROC(glGenProgramsARB);
  if (!glProgramStringARB) GET_PROC(glProgramStringARB);

  /* generate program and read it in to the hardware */
  glGenProgramsARB(1, (GLuint *) &prog_id);
  if(!prog_id) {
    free(source);
    gpubench_setError(GPUBENCH_ERROR_SHADER_LOAD, "GenPrograms failed");
    return GPUBENCH_BOGUS_PROGRAM;
  }

  /* bind program */
  if (!glBindProgramARB) GET_PROC(glBindProgramARB);
  glBindProgramARB(shader_type, (GLuint) prog_id);

  /* read it in... */
  prog_len = strlen(source);
  glProgramStringARB(shader_type, GL_PROGRAM_FORMAT_ASCII_ARB, prog_len, (const GLubyte*) source);


  /* check if there were any errors with the fragment program */
  if (glGetError() != GL_NO_ERROR) {
    GLint ePos;
    char* error = (char*) glGetString(GL_PROGRAM_ERROR_STRING_ARB);
    glGetIntegerv(GL_PROGRAM_ERROR_POSITION_ARB, &ePos);
    fprintf(stderr, "Shader error: %s\n", error);
    fprintf(stderr, "FSHADER::Load(): Error loading the program at location %d: %.30s\n",
            ePos, source + ePos);
    fflush(stderr);

    free(source);
    gpubench_setError(GPUBENCH_ERROR_SHADER_LOAD, "Parse failed");
    return GPUBENCH_BOGUS_PROGRAM;
  }

  /* make sure that the program has been loaded properly */
  if (!glIsProgramARB) GET_PROC(glIsProgramARB);
  if(!glIsProgramARB(prog_id)) {
    free(source);
    gpubench_setError(GPUBENCH_ERROR_SHADER_LOAD, "Not a valid program id.");
    return GPUBENCH_BOGUS_PROGRAM;
  }

  free(source);

  return(prog_id);
}

int load_fragment_program_arb(char const * fname) {
    return load_program_arb(fname, GL_FRAGMENT_PROGRAM_ARB);
}

int load_vertex_program_arb(char const * fname) {
    return load_program_arb(fname, GL_VERTEX_PROGRAM_ARB);
}

void bind_program_arb (shader_t shader, GLenum shader_type) {
  if (!glBindProgramARB) GET_PROC(glBindProgramARB);
  glEnable(shader_type);
  CHECK_GL();
  glBindProgramARB(shader_type, shader);
  CHECK_GL();

  return;
}

void bind_fragment_program_arb(shader_t shader) {
    bind_program_arb(shader, GL_FRAGMENT_PROGRAM_ARB);
}

void bind_vertex_program_arb(shader_t shader) {
    bind_program_arb(shader, GL_VERTEX_PROGRAM_ARB);
}

void release_vertex_program_arb() {
    glDisable(GL_VERTEX_PROGRAM_ARB);
}

void release_fragment_program_arb(void) {
  glDisable(GL_FRAGMENT_PROGRAM_ARB);
  return;
}

void destroy_fragment_program_arb(shader_t shader) {
  if (!glDeleteProgramsARB) GET_PROC(glDeleteProgramsARB);
  glDeleteProgramsARB(1, (GLuint *) &shader);
}

void destroy_vertex_program_arb(shader_t shader) {
    destroy_fragment_program_arb(shader);
}




/* sets a fragment variable.  i have to specify the number of the parameter,
   not the name as is done in the nvidia implementation */
void set_variable (shader_t shader, int num,
                            float x, float y, float z, float w, GLenum shader_type) {
  if (!glProgramEnvParameter4fARB) GET_PROC(glProgramEnvParameter4fARB);
  glProgramEnvParameter4fARB(GL_FRAGMENT_PROGRAM_ARB, num, x, y, z, w);
  CHECK_GL();

  return;
}

void set_fragment_variable (shader_t shader, int num,
                            float x, float y, float z, float w) {
    set_variable(shader, num, x, y, z, w, GL_FRAGMENT_PROGRAM_ARB);
}

void set_vertex_variable (shader_t shader, int num,
                            float x, float y, float z, float w) {
    set_variable(shader, num, x, y, z, w, GL_VERTEX_PROGRAM_ARB);
}


void set_active_texture (GLenum texture){
  if (!glActiveTextureARB) GET_PROC(glActiveTextureARB);
  glActiveTextureARB(texture);
  return;
}
