
#ifdef WIN32
#include <windows.h>
#endif

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include <GL/gl.h>


#include "gpubench.h"

static PFNGLACTIVETEXTUREARBPROC     glActiveTextureARB;

static int gpubench_arch = GPUBENCH_ARCH_UNKNOWN;
static int gpubench_status = GPUBENCH_ERROR_NONE;


static int gpubench_exitOnError = 0;
static int gpubench_verboseWithErrors = 1;

static char gpubench_errorStr[GPUBENCH_STR_LENGTH+1];


// initializes gpubench for detected architecture
static int gpubench_detectArch() {

  const char* vendor;

  if (gpubench_arch != GPUBENCH_ARCH_UNKNOWN)
    return gpubench_arch;

  strcpy(gpubench_errorStr, "");

  vendor = (const char *) glGetString(GL_VENDOR);

  if (vendor == NULL) {
    fprintf(stderr, "NULL vendor string.  Is there a window? Create gl window first.\n");
    exit(1);
  }

  if (strstr(vendor, "NVIDIA") != NULL)
    gpubench_arch = GPUBENCH_ARCH_NV;
  else if (strstr(vendor, "ATI") != NULL)
    gpubench_arch = GPUBENCH_ARCH_ATI;
  else if (strstr(vendor, "3Dlabs") != NULL)
    gpubench_arch = GPUBENCH_ARCH_3DLABS;
  else {
    fprintf(stderr, "**WARNING**: Unknown GPU vendor: '%s'\n", vendor);
  }

  return gpubench_arch;
}


int gpubench_getArch() {
  return gpubench_detectArch();
}

int gpubench_isATI() {
  return (gpubench_detectArch() == GPUBENCH_ARCH_ATI);
}

int gpubench_isNV() {
  return (gpubench_detectArch() == GPUBENCH_ARCH_NV);
}

int gpubench_is3Dlabs() {
  return (gpubench_detectArch() == GPUBENCH_ARCH_3DLABS);
}



const char *gpubench_getDriverVersion(void)
{

  int infosize;
  void *info;
  int status;
  int vsinfosize;
  char *desc;
  LPVOID version=NULL;
  DWORD vLen,langD;
  static char fileVersion[256];
  static char returnValue[256];
  char *p;

  static const char nvdriver[] = "c:\\windows\\system32\\nv4_disp.dll";
  static const char atidriver[] = "c:\\windows\\system32\\atioglxx.dll";
  static const char driver3dl[] = "c:\\windows\\system32\\3Dl2DD.dll";

  const char *filename;

  if (gpubench_isATI()) filename = atidriver;
  else if (gpubench_isNV()) filename = nvdriver;
  else filename = driver3dl;

  sprintf (fileVersion, "");

  infosize = GetFileVersionInfoSize( filename, (LPDWORD) &infosize );
  if (!infosize) return "";

  info = malloc (infosize);

  status = GetFileVersionInfo(filename, 0, infosize, info);
  if (!status) {free(info); return "";}

  status = VerQueryValue ( info, TEXT("\\VarFileInfo\\Translation"),
                           (LPVOID *) &version, (UINT *)&vLen);
  if (!status) {free(info); return "";}

  memcpy(&langD,version,4);
  sprintf(fileVersion, "\\StringFileInfo\\%02X%02X%02X%02X\\FileVersion",
          (langD & 0xff00)>>8,langD & 0xff,(langD & 0xff000000)>>24,
          (langD & 0xff0000)>>16);

  if (VerQueryValue(info, fileVersion, (LPVOID *)&desc, (PUINT) &vsinfosize)) {
     p = desc;
     if (!gpubench_is3Dlabs()) {
        for  (   ; *p && *p != '.'; p++);
        for  (p++; *p && *p != '.'; p++);
        for  (p++; *p && *p != '.'; p++);
        p++;
     } else {
        while (*p != '\0' && *p != ' ') p++;
        if (*p == ' ') *p = '\0';
        p = desc;
     }
  } else {
     p = "";
  }
  strncpy(returnValue, p, sizeof returnValue);
  returnValue[sizeof returnValue - 1] = '\0';

  free (info);
  return returnValue;
}

double gpubench_gigaInstrPerSec(double total_instr, double msec) {
  return total_instr / 1000.0 / 1000.0 / 1000.0 / (msec / 1000.0);
}


double gpubench_gigabytesPerSec(double total_bytes, double msec) {
  return  total_bytes / 1024.0 / 1024.0 / 1024.0 / (msec / 1000.0);
}


int gpubench_fequals(double a, double b) {
  // VERY VERY lax comparison due to ATI precision
  if (fabs(a-b) >= .075f)
    return 0;
  else
    return 1;
}


void gpubench_printFrameBuffer(int minx, int miny, int sizex, int sizey, int components) {

  int i,j,k;
  float* result = (float*)malloc( sizeof(float) * components * sizex * sizey);

  glFinish();

  // KAYVON: bug.  ATI platform readback of 1,2 components.  NV platform using 2 components
  int format[5] = {0, GL_RED, GL_LUMINANCE_ALPHA,
                   GL_RGB, GL_RGBA};

  glReadPixels(minx, miny, sizex, sizey, format[components], GL_FLOAT, result);

  for (i=0;i<sizey;i++) {
    for (j=0;j<sizex;j++) {
      printf("(%d,%d) = ", j, i);
      for (k=0;k<components;k++)
        printf("%.4f ", result[components*(i*sizex + j)+k]);
      printf("\n");
    }
  }

  free(result);
}



int gpubench_verifyFrameBuffer(float* correct, int minx, int miny,
                               int sizex, int sizey, int verbose) {

  // KAYVON: bug. not multicomponent safe
  int i, x, y;
  float* result;
  int failed = FALSE;

  assert(correct);

  result = (float*)malloc( 4 * sizeof(float) * sizex * sizey);
  glFinish();
  glReadPixels(minx, miny, sizex, sizey, GL_RGBA, GL_FLOAT, result);

  for (i=0;i<4*sizex*sizey;i++) {
    if (!gpubench_fequals(correct[i], result[i])) {
      y = (i/4) / sizex;
      x = (i/4) % sizex;

      if (verbose)
        printf("(%d,%d)[%d] exp %f got %f\n", x, y, i%4, correct[i], result[i]);

      failed = 1;
    }
  }

  free(result);
  return failed;
}


void gpubench_finishForReal() {
  // KAYVON: not multi-component safe
  float f[4];
  glFinish();
  glReadPixels(0, 0, 1, 1, GL_RGBA, GL_FLOAT, f);
}


/* shader handling wrappers, should be done with function pointers, I know */

shader_t gpubench_loadShaderFromString(const char *fname) {
  return load_fragment_program_string_arb(fname);
}

shader_t gpubench_loadVertexShaderFromString(char const * fname) {
    return load_vertex_program_string_arb(fname);
}


shader_t gpubench_loadShader(const char *fname) {
  return load_fragment_program_arb(fname);
}

shader_t gpubench_loadVertexShader(char const * fname)
{
    return load_vertex_program_arb(fname);
}


void gpubench_destroyShader(shader_t shader) {
  destroy_fragment_program_arb(shader);
}

void gpubench_destroyVertexShader(shader_t shader) {
    destroy_vertex_program_arb(shader);
}


void gpubench_bindShader(shader_t shader) {
  bind_fragment_program_arb(shader);
}

void gpubench_bindVertexShader(shader_t shader) {
    bind_vertex_program_arb(shader);
}


void gpubench_releaseShader() {
  release_fragment_program_arb();
}

void gpubench_releaseVertexShader() {
    release_vertex_program_arb();
}


void gpubench_setShaderConst(shader_t shader, int num,
                             float x, float y, float z, float w) {
  set_fragment_variable(shader, num, x, y, z, w);
}

void gpubench_setVertexShaderConst(shader_t shader, int num,
                             float x, float y, float z, float w) {
  set_vertex_variable(shader, num, x, y, z, w);
}


static char error_txt[][32] = {
  "GL_INVALID_ENUM",
  "GL_INVALID_VALUE",
  "GL_INVALID_OPERATION",
  "GL_STACK_OVERFLOW",
  "GL_STACK_UNDERFLOW",
  "GL_OUT_OF_MEMORY"
};


bool gpubench_checkGL(int line, char *file)
{
  GLenum r = glGetError();
  if (r	!= GL_NO_ERROR)	{
    fprintf(stderr, "GLError: %s at %s:%d\n",
            error_txt[r - GL_INVALID_ENUM], file, line);
    return false;
  }
  return true;
}


void gpubench_prettyPrintErrors() {

  GLenum r = glGetError();

  fprintf(stderr, "GPUBench Error detected: *********************************\n");
  fprintf(stderr, "GPUBench error is: (%u) -- %s.\n", gpubench_status, gpubench_errorStr);

  if (r != GL_NO_ERROR)
    fprintf(stderr, "OpenGL error is: %d.\n", error_txt[r-GL_INVALID_ENUM]);

  fprintf(stderr, "**********************************************************\n");
}


int gpubench_checkError() {

  return gpubench_status;

}

void gpubench_setError(int errorCode, char* errorStr) {
  gpubench_status = errorCode;
  strncpy(gpubench_errorStr, errorStr, GPUBENCH_STR_LENGTH);

  if (gpubench_exitOnError) {
    gpubench_prettyPrintErrors();
    fprintf(stderr, "ExitOnError enabled, exiting...\n");
    exit(gpubench_status);
  }
  else if (gpubench_verboseWithErrors) {
    gpubench_prettyPrintErrors();
  }
}



float gpubench_render(int size,
                      int winSize,
                      int chunksize,
                      int renderTriangleNotQuad,
                      int iterations,
                      int accessType)
{


  int i, j, k;

  float foffset;

  if ( (chunksize != 0) && (size % chunksize != 0) ) {

    fprintf(stderr, "Error: Chunksize should be multiple of render target size, not running test.\n");
    return 0.0f;
  }

  int wasNotEnabled = glIsEnabled(GL_SCISSOR_TEST);

  glViewport (0, 0, size, size);
  glEnable(GL_SCISSOR_TEST);
  glScissor( 0, 0, size, size);

  int tex = (accessType == GPUBENCH_ACCESS_SEQ) ? 1 : 0;

  foffset = 2.0f * winSize / size - 1.0f;

  float scaleX = (gpubench_isNV()) ? 1.0f : (float)size;
  float scaleY = (gpubench_isNV()) ? 1.0f : (float)size;


  /* force load of all texture data onto card */
  glBegin(GL_QUADS);
  glVertex2f (-1.0f, -1.0f);
  glVertex2f (-1.0f, 1.0f);
  glVertex2f (1.0f, 1.0f);
  glVertex2f (1.0f, -1.0f);
  glEnd();

  glFinish();


  gpubench_resetTimer();

  if (renderTriangleNotQuad) {

    if (chunksize) {

      for (i=0; i<iterations; i++) {
        for (j=0; j<size; j+=chunksize) {
          for (k=0; k<size; k+=chunksize) {

            glScissor(k, j, chunksize, chunksize);

            glBegin(GL_TRIANGLES);

            glTexCoord2f ( 0.0f, 0.0f );
            glVertex2f   (-1.0f, -1.0f);

            glTexCoord2f ( (float) (size*2)*tex/scaleX, 0.0f);
            glVertex2f   (3.0f, -1.0f);

            glTexCoord2f ( 0.0f, (float) (size*2)*tex/scaleY );
            glVertex2f   (-1.0f, 3.0f);

            glEnd();
          }
        }
      }

    } else {

      for (i=0; i<iterations; i++) {

        glBegin(GL_TRIANGLES);

        glTexCoord2f (0.0f, 0.0f);
        glVertex2f   (-1.0f, -1.0f);

        glTexCoord2f ((float) (size*2)*tex/scaleX, 0.0f);
        glVertex2f   (3.0f, -1.0f);

        glTexCoord2f (0.0f, (float) (size*2)*tex/scaleY);
        glVertex2f   (-1.0f, 3.0f);

        glEnd();
      }
    }

  } else {

    if (chunksize) {
      float f = 2.0f / size;

      for (i=0; i<iterations; i++) {
        glBegin(GL_QUADS);
        for (j=0; j<size; j+=chunksize) {
          for (k=0; k<size; k+=chunksize) {

            glScissor(k, j, chunksize, chunksize);

            glTexCoord2f (0.0f, 0.0f);
            glVertex2f   (-1.0f, -1.0f);

            glTexCoord2f ((float) (size*tex)/scaleX, 0.0f);
            glVertex2f   (1.0f,  -1.0f);

            glTexCoord2f ((float) size*tex, (float) (size*tex)/scaleY);
            glVertex2f   (1.0f,   1.0f);

            glTexCoord2f (0.0f, (float) (size*tex)/scaleY);
            glVertex2f   (-1.0f,  1.0f);

          }
        }
        glEnd();

      }
    } else {
      for (i=0; i<iterations; i++) {

        glBegin(GL_QUADS);
        glTexCoord2f (0.0f, 0.0f);
        glVertex2f   (-1.0f, -1.0f);

        glTexCoord2f ((float) (size*tex)/scaleX, 0.0f);
        glVertex2f   (1.0f,  -1.0f);

        glTexCoord2f ((float) size*tex/scaleX, (float) (size*tex)/scaleY);
        glVertex2f   (1.0f,   1.0f);

        glTexCoord2f (0.0f, (float) (size*tex)/scaleY);
        glVertex2f   (-1.0f,  1.0f);
        glEnd();
      }
    }
  }

  glFinish();

  if (wasNotEnabled)
    glDisable(GL_SCISSOR_TEST);

  return gpubench_getTimer();

}


void gpubench_setFlag(int flag, int value) {

  switch (flag) {
    case GPUBENCH_FLAG_EXITONERROR:
      gpubench_exitOnError = value;
      break;
    case GPUBENCH_FLAG_VERBOSE:
      gpubench_verboseWithErrors = value;
      break;
    default:
      gpubench_setError(GPUBENCH_ERROR_INVALID_FLAG, "");
  }
}


char* gpubench_getErrorString() {
  return gpubench_errorStr;
}



GLuint
gpubench_makeTex(int texNum,
                 int width, int height, int nComponents, float *texData)
{
   static int nvtype[5] = {
      0, GL_FLOAT_R32_NV, GL_FLOAT_RG32_NV,
      GL_FLOAT_RGB32_NV, GL_FLOAT_RGBA32_NV
   };
   static int atitype[5] = {
      0, GL_ALPHA_FLOAT32_ATI, GL_LUMINANCE_ALPHA_FLOAT32_ATI,
      GL_RGB_FLOAT32_ATI, GL_RGBA_FLOAT32_ATI
   };
   static int format[5] = { 0, GL_RED, GL_LUMINANCE_ALPHA, GL_RGB, GL_RGBA };
   static bool loadedSyms;
   GLuint texID;

   if (!loadedSyms) {
      loadedSyms = true;
      GETPROCADDR(PFNGLACTIVETEXTUREARBPROC, glActiveTextureARB);
   }
    // note: right now just using 2D textures on ATI since I don't
    // know how to read from RECT render texture on ATI.  This is 
    // unfortunate since we now have to create two shaders (one for 2D
    // textures and one for RECT textures)
    int textureType = gpubench_isNV() ? GL_TEXTURE_RECTANGLE_NV : GL_TEXTURE_2D;

   glEnable(textureType);
   glGenTextures(1, &texID);
   glActiveTextureARB(GL_TEXTURE0_ARB + texNum);
   glBindTexture(textureType, texID);
   glTexParameterf(textureType, GL_TEXTURE_WRAP_S, GL_CLAMP);
   glTexParameterf(textureType, GL_TEXTURE_WRAP_T, GL_CLAMP);
   glTexParameterf(textureType, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
   glTexParameterf(textureType, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
   CHECK_GL();

   glTexImage2D(textureType, 0,
                gpubench_isNV() ? nvtype[nComponents] : atitype[nComponents],
                width, height, 0, format[nComponents], GL_FLOAT, texData);
   CHECK_GL();
   return texID;
}

void gpubench_sendToGPU(int texID, int width, int height, int nComponents, float * texData)
{
   static int nvtype[5] = {
      0, GL_FLOAT_R32_NV, GL_FLOAT_RG32_NV,
      GL_FLOAT_RGB32_NV, GL_FLOAT_RGBA32_NV
   };
   static int atitype[5] = {
      0, GL_ALPHA_FLOAT32_ATI, GL_LUMINANCE_ALPHA_FLOAT32_ATI,
      GL_RGB_FLOAT32_ATI, GL_RGBA_FLOAT32_ATI
   };
   static int format[5] = { 0, GL_RED, GL_LUMINANCE_ALPHA, GL_RGB, GL_RGBA };
   static int textureType = gpubench_isNV() ? GL_TEXTURE_RECTANGLE_NV : GL_TEXTURE_2D;

   glBindTexture(textureType, texID);
   glTexSubImage2D(textureType, 0, 0, 0, width, height, format[nComponents],
                GL_FLOAT, texData);
   CHECK_GL();
}


void
gpubench_freeTex(GLuint texID)
{
   glDeleteTextures(1, &texID);
   CHECK_GL();
}
