#ifndef __GPUBENCH_H__
#define __GPUBENCH_H__

#ifdef _WIN32
#include <windows.h>
#endif

#include <assert.h>

#include <GL/gl.h>

#include "gpubench_enum.h"

#include "nv30glext.h"

#define GETPROCADDR(type, name) {                                          \
      name = (type) wglGetProcAddress(#name);                              \
      if (!name) {fprintf(stderr, "Could Not Load " #name "\n"); exit(1);} \
   }


#define MIN(x,y) (((x) < (y)) ? (x) : (y))
#define MAX(x,y) (((x) > (y)) ? (x) : (y))



typedef int shader_t;

/* gpubench "API" calls ***************************************/

int     gpubench_getArch();
int     gpubench_isATI();
int     gpubench_isNV();
int     gpubench_is3Dlabs();

const char *gpubench_getDriverVersion(void);

void    gpubench_resetTimer();
float   gpubench_getTimer();

void    gpubench_finishForReal();

/* Gigaflops and gigabytes per second calculations
	Note: params are doubles because often overflow integer type with
				intruction counts
*/
double  gpubench_gigaInstrPerSec(double total_instr, double msec);
double  gpubench_gigabytesPerSec(double total_bytes, double msec);

/* forgiving (inexact) float equality test (due to precision round off) */
int     gpubench_fequals(double a, double b);

/* framebuffer inspection */
void    gpubench_printFrameBuffer(int minx, int miny, int sizex, int sizey,
                                  int components);
int     gpubench_verifyFrameBuffer(float* correct, int xmin, int ymin,
                                   int xsize, int ysize, int verbose);

/* generic render big primitive method */
float   gpubench_render(int size, int winSize, int chunksize,
                        int renderTriangleNotQuad, int iterations,
                        int zeroCoord);

/* shader manipulation functions */
shader_t  gpubench_loadShaderFromString(const char *fname);
shader_t  gpubench_loadShader(const char *fname);
void      gpubench_destroyShader(shader_t shader);
void      gpubench_bindShader(shader_t shader);
void      gpubench_releaseShader(void);
void      gpubench_setShaderConst(shader_t shader, int num,
                                  float x, float y, float z, float w);

/* vertex shader manipulation functions */
shader_t  gpubench_loadVertexShaderFromString(const char *fname);
shader_t  gpubench_loadVertexShader(const char *fname);
void      gpubench_destroyVertexShader(shader_t shader);
void      gpubench_bindVertexShader(shader_t shader);
void      gpubench_releaseVertexShader(void);
void      gpubench_setVertexShaderConst(shader_t shader, int num,
                                  float x, float y, float z, float w);

void      gpubench_getprogramiv(GLenum p, GLenum e, int *i);

/* Float texture creation and destruction */
GLuint gpubench_makeTex(int texNum,
                     int width, int height, int nComponents, float *texData);
void gpubench_freeTex(GLuint texID);

void gpubench_sendToGPU(int texID, int width, int height, int nComponents, float * texData);


/* added on 6/24/04, but not completely functional yet */
int       gpubench_checkError();
char*     gpubench_getErrorString();
void      gpubench_setError(int errorCode, char* errorStr);
void      gpubench_setFlag(int flag, int value);

void  gpubench_createWindow(int w, int h, void (*drawfn) (void), char name[]);
int   gpubench_createFloatPbuffer(int w, int h, int doublebuffer);
int   gpubench_createFloatPbufferComponents(int w, int h,
                                            int doublebuffer, int components);
int   gpubench_createFloatPbufferAux(int w, int h,
                                     int doublebuffer, int auxBuffers);
int   gpubench_createFloatPbufferAuxComponents(int w, int h, int doublebuffer,
                                               int auxBuffers, int components);
int   gpubench_createFloatPbufferAll(int w, int h, int doublebuffer,
                                     int auxBuffers, int components, int depth);
void  gpubench_bindPbufferAux(int which);
void  gpubench_releasePbufferAux(int which);
void  gpubench_destroyPbuffer();


void  gpubench_swapBuffers(void);

void  gpubench_makeCurrentPbuffer(void);
void  gpubench_makeCurrentWindow(void);

void  gpubench_showWindow();
void  gpubench_hideWindow();

void  gpubench_bindPbuffer(int front_or_back);
void  gpubench_releasePbuffer(int front_or_back);

int   gpubench_bindAndSwap(int draw_to_front);


#define CHECK_GL() \
   do { bool _r = gpubench_checkGL(__LINE__, __FILE__); assert(_r); } while(0)
#define CHECK_GL_NO_ASSERT() gpubench_checkGL(__LINE__, __FILE__)
bool gpubench_checkGL(int line, char *file);


/*
 * These are actually the private functions used internally between
 * gpubench.cpp and its helper backend files.
 */

shader_t  load_fragment_program_string_arb(const char *fname);
shader_t  load_fragment_program_arb(const char *fname);
void      bind_fragment_program_arb(shader_t shader);
void      release_fragment_program_arb(void);
void      destroy_fragment_program_arb(shader_t shader);
void      set_fragment_variable(shader_t shader, int num,
                                float x, float y, float z, float w);

shader_t  load_vertex_program_string_arb(const char *fname);
shader_t  load_vertex_program_arb(const char *fname);
void      bind_vertex_program_arb(shader_t shader);
void      release_vertex_program_arb(void);
void      destroy_vertex_program_arb(shader_t shader);
void      set_vertex_variable(shader_t shader, int num,
                                float x, float y, float z, float w);


#endif
